import Foundation
import SQLite
import CoreLocation
import HealthKit

typealias Database = Connection

final class HealthDatabase: ObservableObject {
    
    private let fileUrl: URL

    private let database: Connection

    @Published
    var workouts: [Workout] = []
    
    convenience init(fileUrl: URL) throws {
        let database = try Connection(fileUrl.path)
        self.init(fileUrl: fileUrl, database: database)
    }

    init(fileUrl: URL, database: Connection) {
        self.fileUrl = fileUrl
        self.database = database
        DispatchQueue.global().async {
            self.readAllWorkouts()
        }
    }

    func readAllWorkouts() {
        let workouts: [Workout]
        do {
            workouts = try Workout.readAll(in: database)
        } catch {
            print("Failed to read workouts: \(error)")
            return
        }
        DispatchQueue.main.async {
            self.workouts = workouts
        }
    }

    func locationSamples(for activity: HKWorkoutActivity) throws -> [LocationSample] {
        try LocationSample.locationSamples(from: activity.startDate, to: activity.currentEndDate, in: database)
    }

    func locationSampleCount(for activity: HKWorkoutActivity) throws -> Int {
        try LocationSample.locationSampleCount(from: activity.startDate, to: activity.currentEndDate, in: database)
    }

    func samples(for activity: HKWorkoutActivity) throws -> [Sample.DataType : [Sample]] {
        try Sample.samples(from: activity.startDate, to: activity.currentEndDate, in: database).reduce(into: [:]) {
            $0[$1.dataType] = ($0[$1.dataType] ?? []) + [$1]
        }
    }

    func sampleCount(for activity: HKWorkoutActivity) throws -> Int {
        try Sample.sampleCount(from: activity.startDate, to: activity.currentEndDate, in: database)
    }

    var activities: [HKWorkoutActivity] {
        workouts.map { $0.activities }.joined().sorted()
    }

    private func testActivityOverlap() {
        let activities = self.activities
        var current = activities.first!
        for next in activities.dropFirst() {
            let overlap = next.startDate.timeIntervalSince(current.currentEndDate)
            if overlap < 0 {
                print("Overlap \(-overlap.roundedInt) s:")
                print("    Activity \(current.workoutConfiguration.activityType.description): \(current.startDate.timeAndDateText) -> \(current.currentEndDate.timeAndDateText)")
                print("    Activity \(next.workoutConfiguration.activityType.description): \(next.startDate.timeAndDateText) -> \(next.currentEndDate.timeAndDateText)")
            }
            current = next
        }
    }

    convenience init(database: Database) {
        self.init(fileUrl: .init(filePath: "/"), database: database)
    }
}

private extension HKWorkoutActivity {

    var currentEndDate: Date {
        endDate ?? Date()
    }
}
