import Foundation
import SQLite
import HealthKit

extension HKWorkoutActivity: Comparable {

    public static func < (lhs: HKWorkoutActivity, rhs: HKWorkoutActivity) -> Bool {
        lhs.startDate < rhs.startDate
    }
}

enum WorkoutActivityTable {

    private static let table = Table("workout_activities")

    private static let columnId = Expression<Int>("ROWID")

    private static let columnUUID = Expression<Data>("uuid")

    private static let columnOwnerId = Expression<Int>("owner_id")

    private static let columnIsPrimaryActivity = Expression<Bool>("is_primary_activity")

    private static let columnActivityType = Expression<Int>("activity_type")

    private static let columnLocationType = Expression<Int>("location_type")

    private static let columnSwimmingLocationType = Expression<Int>("swimming_location_type")

    private static let columnLapLength = Expression<Data?>("lap_length")

    private static let columnStartDate = Expression<Double>("start_date")

    private static let columnEndDate = Expression<Double>("end_date")

    private static let columnDuration = Expression<Double>("duration")

    private static let columnMetadata = Expression<Data?>("metadata")

    private static func readAll(in database: Connection) throws -> [HKWorkoutActivity] {
        try database.prepare(table).map(activity)
    }

    private static func activity(from row: Row) throws -> HKWorkoutActivity {
        let configuration = HKWorkoutConfiguration()
        configuration.lapLength = try row[columnLapLength].map(lapLength)
        configuration.activityType = .init(rawValue: UInt(row[columnActivityType]))!
        configuration.locationType = .init(rawValue: row[columnLocationType])!
        configuration.swimmingLocationType = .init(rawValue: row[columnSwimmingLocationType])!

        let start = Date(timeIntervalSinceReferenceDate: row[columnStartDate])
        let end = Date(timeIntervalSinceReferenceDate: row[columnEndDate])
        let uuid = row[columnUUID].uuidString
        let metadata: [String : Any] = [HKMetadataKeyExternalUUID : uuid]
        // duration: row[columnDuration]
        // isPrimaryActivity: row[columnIsPrimaryActivity]

        // metadata: row[columnMetadata]
        #warning("Decode metadata")
        return .init(
            workoutConfiguration: configuration,
            start: start,
            end: end,
            metadata: metadata)
    }

    static func activities(for workoutId: Int, in database: Connection) throws -> [HKWorkoutActivity] {
        try database.prepare(table.filter(columnOwnerId == workoutId)).map(activity)
    }

    static func create(in database: Connection) throws {
        //try database.execute("CREATE TABLE workout_activities (ROWID INTEGER PRIMARY KEY AUTOINCREMENT, uuid BLOB UNIQUE NOT NULL, owner_id INTEGER NOT NULL REFERENCES workouts(data_id) ON DELETE CASCADE, is_primary_activity INTEGER NOT NULL, activity_type INTEGER NOT NULL, location_type INTEGER NOT NULL, swimming_location_type INTEGER NOT NULL, lap_length BLOB, start_date REAL NOT NULL, end_date REAL NOT NULL, duration REAL NOT NULL, metadata BLOB)")
        try database.run(table.create { t in
            t.column(columnId, primaryKey: .autoincrement)
            t.column(columnUUID)
            t.column(columnOwnerId, references: Table("workouts"), Expression<Int>("data_id"))
            t.column(columnIsPrimaryActivity)
            t.column(columnActivityType)
            t.column(columnLocationType)
            t.column(columnSwimmingLocationType)
            t.column(columnLapLength)
            t.column(columnStartDate)
            t.column(columnEndDate)
            t.column(columnDuration)
            t.column(columnMetadata)
        })
    }

    static func insert(_ element: HKWorkoutActivity, isPrimaryActivity: Bool, dataId: Int, in database: Connection) throws {
        try database.run(table.insert(
            columnUUID <- (element.externalUUID ?? element.uuid).uuidString.data(using: .utf8)!,
            columnOwnerId <- dataId,
            columnIsPrimaryActivity <- isPrimaryActivity,
            columnActivityType <- Int(element.workoutConfiguration.activityType.rawValue),
            columnLocationType <- element.workoutConfiguration.locationType.rawValue,
            columnSwimmingLocationType <- element.workoutConfiguration.swimmingLocationType.rawValue,
            columnLapLength <- try lapLengthData(lapLength: element.workoutConfiguration.lapLength),
            columnStartDate <- element.startDate.timeIntervalSinceReferenceDate,
            columnEndDate <- element.endDate?.timeIntervalSinceReferenceDate ?? element.startDate.addingTimeInterval(element.duration).timeIntervalSinceReferenceDate,
            columnDuration <- element.duration)
            //columnMetadata <- element.metadata)
        )
    }
}

private extension HKWorkoutActivity {

    var externalUUID: UUID? {
        guard let string = metadata?[HKMetadataKeyExternalUUID] as? String else {
            return nil
        }
        return UUID(uuidString: string)
    }
}

private extension WorkoutActivityTable {

    static func lapLengthData(lapLength: HKQuantity?) throws -> Data? {
        try lapLength.map { try NSKeyedArchiver.archivedData(withRootObject: $0, requiringSecureCoding: false) }
    }

    static func lapLength(from data: Data) throws -> HKQuantity? {
        try NSKeyedUnarchiver.unarchivedObject(ofClass: HKQuantity.self, from: data)
    }
}

private extension Data {

    var uuidString: String {
        let h = Array(self)
        let parts = [h[0..<4], h[4..<6], h[6..<8], h[8..<10], h[10..<16]]
        return parts.map { Data($0).hex }.joined(separator: "-")
    }
}

private extension UUID {

    init?(data: Data) {
        self.init(uuidString: data.uuidString)
    }
}
